#  Bitcoin Address  v0.1
#  Copyright (c) 2020 - https://github.com/fortesp/bitcoinaddress
#  This software is distributed under the terms of the MIT License.
#  See the file 'LICENSE' in the root directory of the present distribution,
#  or http://opensource.org/licenses/MIT.
import hashlib
from abc import ABC
import base58
import binascii
from bitcoinaddress.key.key import Key
from . import segwit_addr
from .util import hash160, ecdsa_secp256k1, checksum


class Address:
    PREFIX_EVEN = b'\x02'
    PREFIX_ODD = b'\x03'
    WITNESS_VERSION = 0x00

    class Net(ABC):
        def __init__(self):
            self.pubaddr1 = None
            self.pubaddr1c = None
            self.pubaddr3 = None
            self.pubaddrbc1_P2WPKH = None
            self.pubaddrbc1_P2WSH = None

    class MainNet(Net):
        PREFIX_A = b'\x00'
        PREFIX_B = b'\x04'
        SEGWIT_PREFIX = b'\x05'
        BECH32_PREFIX = "bc"

        def __init__(self, instance):
            self.instance = instance

        def generate_publicaddress1(self):
            self.pubaddr1 = self.instance._generate_publicaddress1(Address.MainNet.PREFIX_A,
                                                                   Address.MainNet.PREFIX_B)

        def generate_publicaddress1_compressed(self):
            self.pubaddr1c = self.instance._generate_publicaddress1_compressed(Address.MainNet.PREFIX_A)

        def generate_publicaddress3(self):
            self.pubaddr3 = self.instance._generate_publicaddress3(Address.MainNet.SEGWIT_PREFIX)

        def generate_publicaddress_bc1_P2WPKH(self):
            self.pubaddrbc1_P2WPKH = self.instance._generate_publicaddress_bech32_P2WPKH(Address.MainNet.BECH32_PREFIX)

        def generate_publicaddress_bc1_P2WSH(self):
            self.pubaddrbc1_P2WSH = self.instance._generate_publicaddress_bech32_P2WSH(Address.MainNet.BECH32_PREFIX)

    class TestNet(Net):
        PREFIX_A = b'\x6F'
        PREFIX_B = b'\x04'
        SEGWIT_PREFIX = b'\xC4'
        BECH32_PREFIX = "tb"

        def __init__(self, instance):
            self.instance = instance

        def generate_publicaddress1(self):
            self.pubaddr1 = self.instance._generate_publicaddress1(Address.TestNet.PREFIX_A,
                                                                   Address.TestNet.PREFIX_B)

        def generate_publicaddress1_compressed(self):
            self.pubaddr1c = self.instance._generate_publicaddress1_compressed(Address.TestNet.PREFIX_A)

        def generate_publicaddress3(self):
            self.pubaddr3 = self.instance._generate_publicaddress3(Address.TestNet.SEGWIT_PREFIX)

        def generate_publicaddress_tb1_P2WPKH(self):
            self.pubaddrtb1_P2WPKH = self.instance._generate_publicaddress_bech32_P2WPKH(Address.TestNet.BECH32_PREFIX)

        def generate_publicaddress_tb1_P2WSH(self):
            self.pubaddrtb1_P2WSH = self.instance._generate_publicaddress_bech32_P2WSH(Address.TestNet.BECH32_PREFIX)

    def __init__(self, key: Key):
        self.key = key
        self.pubkey = None
        self.pubkeyc = None
        self.mainnet = Address.MainNet(self)
        self.testnet = Address.TestNet(self)

    @staticmethod
    def of(key: Key):
        address = Address(key)
        address.generate()
        return address

    def generate(self) -> {}:
        if self.key.digest is None:
            self.key.generate()
        self.mainnet.generate_publicaddress1()
        self.mainnet.generate_publicaddress1_compressed()
        self.mainnet.generate_publicaddress3()
        self.mainnet.generate_publicaddress_bc1_P2WPKH()
        self.mainnet.generate_publicaddress_bc1_P2WSH()
        self.testnet.generate_publicaddress1()
        self.testnet.generate_publicaddress1_compressed()
        self.testnet.generate_publicaddress3()
        self.testnet.generate_publicaddress_tb1_P2WPKH()
        self.testnet.generate_publicaddress_tb1_P2WSH()

    def _generate_publicaddress1(self, prefix_a, prefix_b):
        p = self._generate_uncompressed_pubkey(prefix_b)
        m = prefix_a + hash160(p).digest()
        c = checksum(m)
        return base58.b58encode(m + c).decode('utf-8')

    def _generate_publicaddress1_compressed(self, prefix):
        p = self._generate_compressed_pubkey()
        m = prefix + hash160(p).digest()
        c = checksum(m)
        return base58.b58encode(m + c).decode('utf-8')

    def _generate_publicaddress3(self, prefix):
        prefix_redeem = b'\x00\x14'
        p = self._generate_compressed_pubkey()
        redeem_script = hash160(prefix_redeem + hash160(p).digest()).digest()  # 20 bytes
        m = prefix + redeem_script
        c = checksum(m)
        return base58.b58encode(m + c).decode('utf-8')

    def _generate_publicaddress_bech32_P2WPKH(self, bech32_prefix):
        p = self._generate_compressed_pubkey()
        redeem_script_P2WPKH = hash160(p).digest()  # 20 bytes
        return str(segwit_addr.encode(bech32_prefix, Address.WITNESS_VERSION, redeem_script_P2WPKH))

    def _generate_publicaddress_bech32_P2WSH(self, bech32_prefix):
        p = self._generate_compressed_pubkey()
        redeem_script_P2WSH = hashlib.sha256(p).digest()
        return str(segwit_addr.encode(bech32_prefix, Address.WITNESS_VERSION, redeem_script_P2WSH))

    def _generate_uncompressed_pubkey(self, prefix):
        digest = self.key.digest
        ret = prefix + ecdsa_secp256k1(digest).to_string()  # 1 + 32 bytes + 32 bytes
        self.pubkey = str(binascii.hexlify(ret).decode('utf-8'))
        return ret

    def _generate_compressed_pubkey(self):
        ecdsa_digest = ecdsa_secp256k1(self.key.digest).to_string()
        x_coord = ecdsa_digest[:int(len(ecdsa_digest) / 2)]
        y_coord = ecdsa_digest[int(len(ecdsa_digest) / 2):]
        if int(binascii.hexlify(y_coord), 16) % 2 == 0:
            p = Address.PREFIX_EVEN + x_coord
        else:
            p = Address.PREFIX_ODD + x_coord
        self.pubkeyc = str(binascii.hexlify(p).decode('utf-8'))
        return p

    def __str__(self, testnet=False):
        if testnet:
            params = (self.pubkey,
                      self.pubkeyc,
                      self.testnet.pubaddr1,
                      self.testnet.pubaddr1c,
                      self.testnet.pubaddr3,
                      self.testnet.pubaddrtb1_P2WSH,
                      self.testnet.pubaddrtb1_P2WPKH)
        else:
            params = (self.pubkey,
                      self.pubkeyc,
                      self.mainnet.pubaddr1,
                      self.mainnet.pubaddr1c,
                      self.mainnet.pubaddr3,
                      self.mainnet.pubaddrbc1_P2WSH,
                      self.mainnet.pubaddrbc1_P2WPKH)

        return """Public Key: %s 
                      \rPublic Key compressed: %s\n
                      \rPublic Address 1: %s   
                      \rPublic Address 1 compressed: %s   
                      \rPublic Address 3: %s  
                      \rPublic Address bc1 P2WPKH: %s    
                      \rPublic Address bc1 P2WSH: %s  
                    """ % (params)



#  Bitcoin Address  v0.1
#  Copyright (c) 2020 - https://github.com/fortesp/bitcoinaddress
#  This software is distributed under the terms of the MIT License.
#  See the file 'LICENSE' in the root directory of the present distribution,
#  or http://opensource.org/licenses/MIT.

from bitcoinaddress.key.key import Key
from bitcoinaddress import Address, Seed


class Wallet:

    def __init__(self, hash_or_seed=None, testnet=False):
        if hash_or_seed is None: hash_or_seed = Seed()
        self.key = Key.of(hash_or_seed)
        self.address = Address.of(self.key)
        self.testnet = testnet

    def __str__(self):
        return """%s\n%s""" % (self.key.__str__(self.testnet),
                               self.address.__str__(self.testnet))


if __name__ == "__main__":
    wallet = Wallet()
    print(wallet)



#  Bitcoin Address  v0.1
#  Copyright (c) 2020 - https://github.com/fortesp/bitcoinaddress
#  This software is distributed under the terms of the MIT License.
#  See the file 'LICENSE' in the root directory of the present distribution,
#  or http://opensource.org/licenses/MIT.

import time
import os
from random import randrange


class Seed:

    def __init__(self, entropy=None):
        self.entropy = entropy
        if self.entropy is None:
            self.generate()

    def generate(self):
        self.entropy = Seed.random()

    @staticmethod
    def of(entropy=None):
        return Seed(entropy)

    @staticmethod
    def random():
        # from bitcoin project
        return str(os.urandom(32).hex()) \
               + str(randrange(2 ** 256)) \
               + str(int(time.time() * 1000000))

    def __str__(self):
        return self.entropy



#  Bitcoin Address  v0.1
#  Copyright (c) 2020 - https://github.com/fortesp/bitcoinaddress
#  This software is distributed under the terms of the MIT License.
#  See the file 'LICENSE' in the root directory of the present distribution,
#  or http://opensource.org/licenses/MIT.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bitcoinaddress",
    version="0.1.4",
    author="Pedro Fortes",
    author_email="pedro.daniel.g@gmail.com",
    description="A simple Bitcoin address generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fortesp/bitcoinaddress",
    packages=setuptools.find_packages(),
    install_requires=['base58', 'ecdsa'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



#  Bitcoin Address  v0.1
#  Copyright (c) 2020 - https://github.com/fortesp/bitcoinaddress
#  This software is distributed under the terms of the MIT License.
#  See the file 'LICENSE' in the root directory of the present distribution,
#  or http://opensource.org/licenses/MIT.

import hashlib
from abc import ABC
import base58
from binascii import hexlify, unhexlify
from bitcoinaddress.key.seed import Seed
from bitcoinaddress.util import checksum


class Key:

    class Net(ABC):
        def __init__(self):
            self.wif = None
            self.wifc = None

    class MainNet(Net):
        PREFIX = b'\x80'
        SUFFIX = b'\x01'

        def __init__(self, instance):
            self.instance = instance

        def generate_wif(self):
            self.wif = self.instance._generate_wif(Key.MainNet.PREFIX)

        def generate_wif_compressed(self):
            self.wifc = self.instance._generate_wif_compressed(Key.MainNet.PREFIX, Key.MainNet.SUFFIX)

    class TestNet(Net):
        PREFIX = b'\xEF'
        SUFFIX = b'\x01'

        def __init__(self, instance):
            self.instance = instance

        def generate_wif(self):
            self.wif = self.instance._generate_wif(Key.TestNet.PREFIX)

        def generate_wif_compressed(self):
            self.wifc = self.instance._generate_wif_compressed(Key.TestNet.PREFIX, Key.TestNet.SUFFIX)

    def __init__(self):
        self.seed = None
        self.digest = None
        self.hex = None
        self.mainnet = Key.MainNet(self)
        self.testnet = Key.TestNet(self)

    @staticmethod
    def of(obj):
        key = Key()
        if isinstance(obj, Seed):
            key._from_seed(obj)
        else:
            try:
                if len(obj) == 64:
                    key._from_hex(obj)
                    return key
                if len(obj) == 51:
                    key._from_wif(obj)
                    return key
                if len(obj) == 52:
                    pass  # TODO
            except:
                raise Exception("Unsupported format.")

        return key

    def _from_seed(self, seed: Seed):
        self.seed = seed
        self._generate_digest()
        self._generate_hex()
        self.mainnet.generate_wif()
        self.mainnet.generate_wif_compressed()
        self.testnet.generate_wif()
        self.testnet.generate_wif_compressed()

    def _from_hex(self, hex: str):
        self.hex = hex
        self.digest = unhexlify(hex)
        self.mainnet.generate_wif()
        self.mainnet.generate_wif_compressed()
        self.testnet.generate_wif()
        self.testnet.generate_wif_compressed()

    def _from_wif(self, wif: str):
        checksum_size = 4
        self.digest = base58.b58decode(wif)[1:-checksum_size]
        self._generate_hex()
        self.mainnet.generate_wif()
        self.mainnet.generate_wif_compressed()
        self.testnet.generate_wif()
        self.testnet.generate_wif_compressed()

    def _generate_digest(self):
        entropy = str(self.seed.entropy)
        hash = hashlib.sha256(entropy.encode())
        self.digest = hash.digest()

    def _generate_hex(self):
        self.hex = hexlify(self.digest).decode()

    def _generate_wif(self, prefix):
        digest = prefix + self.digest
        c = checksum(digest)
        return base58.b58encode(digest + c).decode('utf-8')

    def _generate_wif_compressed(self, prefix, suffix):
        digest = prefix + self.digest
        c = checksum(digest + suffix)
        return base58.b58encode(digest + suffix + c).decode('utf-8')

    def __str__(self, testnet=False):
        if testnet:
            params = (self.hex,
                      self.testnet.wif,
                      self.testnet.wifc)
        else:
            params = (self.hex,
                      self.mainnet.wif,
                      self.mainnet.wifc)

        return """
              \rPrivate Key HEX: %s
              \rPrivate Key WIF: %s
              \rPrivate Key WIF compressed: %s 
            """ % (params)



#  Bitcoin Address  v0.1
#  Copyright (c) 2020 - https://github.com/fortesp/bitcoinaddress
#  This software is distributed under the terms of the MIT License.
#  See the file 'LICENSE' in the root directory of the present distribution,
#  or http://opensource.org/licenses/MIT.

import hashlib
import ecdsa


def sha256(v):
    return hashlib.sha256(v)


def doublehash256(v):
    return sha256(sha256(v).digest())


def hash160(v):
    return ripemd(hashlib.sha256(v).digest())


def ripemd(v):
    r = hashlib.new('ripemd160')
    r.update(v)
    return r


def ecdsa_secp256k1(digest):
    sk = ecdsa.SigningKey.from_string(digest, curve=ecdsa.SECP256k1)
    return sk.get_verifying_key()


def checksum(v):
    checksum_size = 4
    return doublehash256(v).digest()[:checksum_size]
